""" base classes. """

__txt__ = "Timestamped based JSON objects."
__version__ = 110

import _thread
import datetime
import importlib
import json
import logging
import os
import os.path
import string
import time
import types

basic_types = [str, int, float, bool, None]
allowedchars = string.ascii_letters + string.digits + "_,-. \n" + string.punctuation
allowednamechars = string.ascii_letters + string.digits + '!.@'
starttime = time.time()

class EOVERLOAD(Exception):
    pass

class ENOPATH(Exception):
    pass

class Object:

    def __iter__(self):
        return iter(self.__dict__)

    def __repr__(self) -> str:
        return "%s.%s at %s" % (
            self.__class__.__module__,
            self.__class__.__name__,
            str(hex(id(self)))
        )

    def __str__(self):
        return json.dumps(self, default=smooth, indent=4, sort_keys=True)

    def __hash__(self):
        return hash(str(self))

    def cache(self, key):
        global cache
        try:
            return cache[key]
        except KeyError:
            cache[key] = self.load(key)
            return cache[key]

    def dumps(self):
        return json.dumps(self, default=smooth, sort_keys=True)

    def get_attr(self, k, d=""):
        return getattr(self, k, d)

    def get_keys(self):
        res = []
        for k in self.__dict__.keys():
            if k.startswith("_"):
                continue
            res.append(k)
        return res

    def load(self, path):
        global cache
        assert path
        assert cfg.workdir
        p = os.path.abspath(os.path.join(cfg.workdir, path))
        with open(p) as f:
            try:
                data = json.load(f)
            except json.decoder.JSONDecodeError:
                logging.error("decode error in %s" % path)
                return self
            for k, v in data.items():
                self.set_attr(k, v)
        self._container = {}
        self._container["path"] = path
        return self

    def search(self, m: None):
        if not m:
            m = {}
        res = False
        for k, v in m.items():
            if k.startswith("_"):
                continue
            vv = self.get_attr(k)
            if v in str(vv):
                res = True
            else:
                res = False
                break
        return res

    def save(self, path="", stime="", timed=True):
        assert cfg.workdir
        if not path and not stime:
            try:
                path = self._container["path"]
            except (TypeError, AttributeError, KeyError):
                pass
        if not path and timed:
            path = os.path.join(get_type(self), stime or rtime())
        if not path:
            txt = str(self)
            raise ENOPATH(txt)
        self._container = {}
        self._container["path"] = path
        p = os.path.abspath(os.path.join(cfg.workdir, path))
        d = os.path.dirname(p)
        if not os.path.exists(d):
            cdir(d)
        with open(p, "w") as f:
            json.dump(self, f, default=smooth, indent=4, sort_keys=True)
        if path not in cache:
            cache[path] = self
        return path

    def set_attr(self, k, v):
        return setattr(self, k, v)

class Dotted(Object, dict):

    def __getattribute__(self, attrname):
        try:
            return self[attrname]
        except KeyError:
            return super().__getattribute__(attrname)

    def __getattr__(self, attrname):
        try:
            return self[attrname]
        except KeyError:
            try:
                return super().__getattribute__(attrname)
            except AttributeError:
                if "_default" in self:
                    self[attrname] = self["_default"]
        return self.__getattribute__(attrname)

    def __setattr__(self, key, value):
        func = self.get(key)
        if isinstance(func, types.MethodType):
            raise EOVERLOAD(key)
        if isinstance(value, types.MethodType):
            return super().__setattr__(key, value)
        return self.__setitem__(key, value)

    def __setitem__(self, key, value):
        func = self.get(key)
        if isinstance(func, types.MethodType):
            raise EOVERLOAD(key)
        return super().__setitem__(key, value)

    def join(self):
        pass

    def format(self, keys=None, full=False):
        if keys is None:
            keys = sorted(self.keys())
        res = []
        txt = ""
        for k in set(keys):
            if k.startswith("_"):
                continue
            try:
                v = self[k]
            except (KeyError, TypeError):
                continue
            if not v:
                continue
            v = str(v)
            if full:
                res.append("%s=%s " % (k, v))
            else:
                res.append(v)
        for v in res:
            txt += "%s " % v.strip()
        return txt.strip()

    def upgrade(self, d):
        for k, v in d.items():
            if v:
                self[k] = v
        return self

    def upgrade2(self, d):
        for k, v in self.items():
            if k in d:
                if not v:
                    self[k] = d[k]

class Log(Dotted):

    pass

class OutputCache(Dotted):

    def add(self, dest, txt):
        if not dest or not txt:
            return
        if "dest" not in self:
            self[dest] = []
        self[dest].append(txt)

    def more(self, dest, nr=10):
        if "dest" in dir(self):
            res = self[dest][-nr:]
            del self[dest][-nr:]
            return res
        return []

class Register(Dotted):

    def register(self, k, v):
        if k not in self:
            self[k] = []
        self[k].append(v)
        return self[k]

class Cfg(Dotted):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._default = ""

    def __iter__(self):
        for k in self.keys():
            if k.startswith("_"):
                continue
            yield k

def cdir(path):
    res = ""
    for p in path.split(os.sep):
        res += "%s%s" % (p, os.sep)
        p = os.path.abspath(os.path.normpath(res))
        try:
            os.mkdir(p)
        except (IsADirectoryError, NotADirectoryError, FileExistsError):
            pass
    return True

def get_type(o):
    return o.__class__.__module__ + "." + o.__class__.__qualname__

def get_types(name=None):
    res = []
    for x in os.listdir(cfg.workdir):
       if name is not None and name.lower() in x.lower():
           res.append(x)
       elif name is None:
           res.append(x)
    return res

def rtime():
    res = str(datetime.datetime.now()).replace(" ", os.sep)
    return res

def smooth(obj):
    if isinstance(obj, Object):
        return vars(obj)
    if type(obj) not in [dict, list, str, int, float, bool, None]:
        return repr(obj)
    return str(obj)

cache = Dotted()
cfg = Cfg(_default="")
cfg.workdir = ""
cfg.logdir = ""
