""" shell related commands. """

import atexit
import logging
import logging.handlers
import optparse
import os
import os.path
import readline
import rlcompleter
import stat
import sys
import termios
import time

from obj.base import Cfg, Dotted, OutputCache, cdir, cfg, get_type,  __version__
from obj.command import Command
from obj.loader import names
from obj.handler import Handler
from obj.store import Store
from obj.tasks import launch
from obj.utils import get_exception, get_name, hd

histfile = ""
resume = {}

opts = [
    ('-a', '', 'store_true', False, 'all', 'load all modules'),
    ('-b', '', 'store_true', False, 'background', 'enable daemon mode.'),
    ('-c', '', "string", '', 'channel', 'channel to join.'),
    ('-d', '', 'string', '', 'workdir', 'set working directory.'),
    ('-l', '', 'string', 'error', 'level', 'loglevel.'),
    ('-m', '', 'string', '', 'modules', 'modules to load.'),
    ('-n', '', 'string', '', 'nick', 'nickname to use when joining a channel'),
    ('-o', '', "string", '', 'options', 'character options to use.'),
    ('-p', '', "string", '', 'password', 'provide password to login with.'),
    ('-s', '', 'string', '', 'server', 'connect server channel nick.'),
    ('-v', '', 'store_true', False, 'verbose', 'enable verbose mode.'),
    ('-w', '', 'store_true', False, 'write', 'use written configuration.'),
    ('-x', '', 'string', '', 'exclude', 'exclude module'),
    ('-y', '', 'store_true', False, 'yes', 'use yes as default answer.'),
    ('-z', '', "store_true", False, 'shell', 'enable shell.'),
    ('', '--debug', "store_true", False, 'debug', 'enable debug mode.'),
    ('', '--logdir', "string", "" , 'logdir', 'logging directory'),
    ('', '--name', "string", "", 'name', 'name of the program.'),
    ('', '--port', "store_true", False, 'port', 'provide port number to connect to.'),
    ('', '--resume', "store_true", False, 'resume', 'resume from filedescriptor.'),
]

class Completer(rlcompleter.Completer):

    def __init__(self, commands=None):
        super().__init__()
        self.commands = commands or []
        self.matches = []

    def complete(self, text, state):
        if state == 0:
            if text: 
                self.matches = [s for s in self.commands if s and s.startswith(text)]
            else:
                self.matches = self.commands[:]
        try:
            return self.matches[state]
        except IndexError:
            return None

class DumpHandler(logging.StreamHandler):

    propagate = False

    def emit(self, record):
        pass

class Shell(Handler):

    def __init__(self):
        super().__init__()
        self._err = sys.stderr
        self._in = sys.stdin
        self._out = sys.stdout
        self._type = get_type(self)
        self._userhosts = Dotted()
        self.prompt = True
        self.verbose = True

    def announce(self, txt):
        self.raw(txt)

    def cmd(self, txt, options="", origin=""):
        if not cfg.workdir:
            cfg.workdir = hd(".%s" % cfg.name)
            cdir(cfg.workdir)
        txt = self.get_aliased(txt)
        e = Command(txt)
        e.orig = repr(self)
        e.origin = origin or "root@shell"
        super().start()
        self.dispatch(e)
        self.ready()
        return e

    def dispatch(self, event):
        super().dispatch(event)
        event.wait()
        self.show_prompt()
     
    def get_event(self):
        txt = input()
        if not txt:
            self._out.write("\n")
            self.show_prompt()
            return
        txt = self.get_aliased(txt)
        e = Command(txt)
        e.orig = repr(self)
        e.origin = "root@shell"
        return e

    def raw(self, txt):
        if self.verbose:
            self._out.write(str(txt))
            self._out.write("\n")

    def resume(self):
        pass

    def say(self, objid, channel, txt):
        self.raw(txt)

    def show_prompt(self):
        if self.prompt:
            self._out.write("> ")
            self._out.flush()
        
    def start(self):
        if not cfg.workdir:
            cfg.workdir = hd(".obj")
            cdir(cfg.workdir)
        if cfg.resume:
            self.resume()
        super().start()
        set_completer(self.names.keys())

    def wait(self):
        if cfg.shell:
            super().wait()
        
def check_permissions(path, dirmask=0o700, filemask=0o600):
    uid = os.getuid()
    gid = os.getgid()
    try:
        stats = os.stat(path)
    except FileNotFoundError:
        return
    except OSError:
        dname = os.path.dirname(path)
        cdir(dname)
        stats = os.stat(dname)
    if stats.st_uid != uid:
        os.chown(path, uid, gid)
    if os.path.isfile(path):
        mask = filemask
    else:
        mask = dirmask
    mode = oct(stat.S_IMODE(stats.st_mode))
    if mode != oct(mask):
        os.chmod(path, mask)

def close_history():
    try:
        readline.write_history_file(histfile)
    except:
        pass

def daemon():
    pid = os.fork()
    if pid != 0:
        reset()
        os._exit(0)
    if not os.path.exists(cfg.logdir):
        cdir(cfg.logdir)
    path = os.path.join(cfg.logdir, "pidfile")
    f = open(path, 'w')
    f.write(str(os.getpid()))
    f.flush()
    f.close()
    os.setsid()
    os.umask(0)
    si = open("/dev/null", 'r')
    so = open("/dev/null", 'a+')
    se = open("/dev/null", 'a+')
    os.dup2(si.fileno(), sys.stdin.fileno())
    os.dup2(so.fileno(), sys.stdout.fileno())
    os.dup2(se.fileno(), sys.stderr.fileno())

def enable_history():
    readline.read_history_file(histfile)
    atexit.register(close_history)

def get_completer() -> Completer:
    return readline.get_completer()

def ld(*args):
    return os.path.abspath(os.path.join(cfg.logdir, *args))

def level(loglevel, logfile="obj.log", nostream=False):
    if not os.path.exists(cfg.logdir):
        cdir(cfg.logdir)
    logfile = os.path.join(cfg.logdir, logfile)
    datefmt = '%H:%M:%S'
    format_time = "%(asctime)-8s %(message)-70s"
    format_plain = "%(message)-0s"
    loglevel = loglevel.upper()
    logger = logging.getLogger("")
    if logger.handlers:
        for handler in logger.handlers:
            logger.removeHandler(handler)
    if logger.handlers:
        for handler in logger.handlers:
            logger.removeHandler(handler)
    try:
        logger.setLevel(loglevel)
    except ValueError:
        print("no level %s" % loglevel)
        return
    formatter = logging.Formatter(format_plain, datefmt)
    if nostream:
        dhandler = DumpHandler()
        dhandler.propagate = False
        dhandler.setLevel(loglevel)
        logger.addHandler(dhandler)
    else:
        handler = logging.StreamHandler()
        handler.propagate = False
        handler.setFormatter(formatter)
        handler.setLevel(loglevel)
        logger.addHandler(handler)
    formatter2 = logging.Formatter(format_time, datefmt)
    filehandler = logging.handlers.TimedRotatingFileHandler(logfile, 'midnight')
    filehandler.propagate = False
    filehandler.setFormatter(formatter2)
    filehandler.setLevel(loglevel)
    logger.addHandler(filehandler)
    return logger

def make_opts(options, usage, version):
    parser = optparse.OptionParser(usage=usage, version=version)
    for opt in options:
        otype, deft, dest, htype = opt[2:]
        if "store" in otype:
            parser.add_option(opt[0], opt[1], action=otype, default=deft, dest=dest, help=htype)
        else:
            parser.add_option(opt[0], opt[1], type=otype, default=deft, dest=dest, help=htype)
    return parser.parse_args()

def parse_cli(name="obj", options=None, usage="", version="", wd=""):
    global histfile
    ver = "%s #%s" % (name.upper(), version or __version__)
    if not options:
        options = opts
    opt, arguments = make_opts(options, usage, ver)
    cfg.update(vars(opt))
    cfg.logdir = os.path.abspath(os.path.join(os.path.expanduser("~"), ".logs"))
    cfg.name = name
    cfg.version = version or __version__
    cfg.args = arguments
    if not cfg.workdir and wd:
        cfg.workdir = wd
    if not cfg.workdir:
        cfg.workdir = os.path.abspath(os.path.join(os.path.expanduser("~"), ".%s" % cfg.name))
    if not os.path.exists(cfg.workdir):
        cdir(cfg.workdir)
    if not os.path.exists(cfg.logdir):
        cdir(cfg.logdir)
    histfile = ld("history")
    touch(histfile)
    readline.read_history_file(histfile)
    return cfg

def reset():
    close_history()
    if "old" in resume:
        termreset(resume["fd"], resume["old"])

def set_completer(commands):
    completer = Completer(commands)
    readline.set_completer(completer.complete)
    readline.parse_and_bind("tab: complete")
    atexit.register(lambda: readline.set_completer(None))

def start(name, version=__version__, wd=""):
    parse_cli(name, opts, "%s [options]" % name, version, wd or hd(".%s" % name.lower()))
    level(cfg.level, "%s.log" % name)
    if not os.path.exists(cfg.workdir):
        cdir(cfg.workdir)
    prev = Cfg(cfg)
    store = Store()
    last = store.last(get_type(cfg))
    if last:
        cfg.update(last)
    if cfg != prev:
        cfg.save()
    shell = Shell()
    shell.start()
    if cfg.args:
        shell.prompt = False
        e = shell.cmd(" ".join(cfg.args))
        e.wait()
        return shell
    if cfg.background:
        daemon()
    logging.warn("%s started at %s" % (name.upper(), time.ctime(time.time())))
    shell.prompt = cfg.shell
    res = []
    thrs = []
    for m in cfg.modules.split(","):
        if not m:
            continue
        try:
            shell.walk(m, True)
        except Exception as ex:
            logging.error(get_exception())
    if cfg.shell:
        shell.show_prompt()
    set_completer(shell.names.keys())
    return shell

def startup():
    resume["fd"] = sys.stdin.fileno()
    resume["old"] = termsetup(sys.stdin.fileno())
    atexit.register(reset)

def termreset(fd, old):
    termios.tcsetattr(fd, termios.TCSADRAIN, old)

def termsetup(fd):
    old = termios.tcgetattr(fd)
    return old

def touch(fname):
    try:
        fd = os.open(fname, os.O_RDONLY | os.O_CREAT)
        os.close(fd)
    except TypeError:
        pass

if __name__ == "__main__":
    startup()
    try:
        shell = Shell()
        shell.start()
        shell.wait()
    except Exception as ex:
        logging.error(get_exception())
    reset()    
